/*
 * This file is part of ce.
 *
 * Copyright (c) 1990 by Chester Ramey.
 *
 * Permission is hereby granted to copy, reproduce, redistribute or
 * otherwise use this software subject to the following: 
 * 	1) That there be no monetary profit gained specifically from 
 *	   the use or reproduction of this software.
 * 	2) This software may not be sold, rented, traded or otherwise 
 *	   marketed.
 * 	3) This copyright notice must be included prominently in any copy
 * 	   made of this software.
 *
 * The author makes no claims as to the fitness or correctness of
 * this software for any use whatsoever, and it is provided as is. 
 * Any use of this software is at the user's own risk.
 */
/*
 * ttyio.c -- Unix terminal I/O.
 *
 * The functions in this file negotiate with the operating system for
 * keyboard characters, and write characters to the display in a barely 
 * buffered fashion.
 *
 * This code works with the 4.2 BSD terminal driver as described in tty(4).
 */

/* order is important here... */
#include	<stdio.h>
#include	"ce.h"
#include	<signal.h>
#include	<sys/types.h>
#include	<sgtty.h>

#define	NOBUF	512			/* Output buffer size.		*/

static char	obuf[NOBUF];		/* Output buffer.		*/
static int	nobuf;			/* # of chars in output buffer	*/
static struct	sgttyb	oldtty;		/* V6/V7 stty data.		*/
static struct	sgttyb	newtty;
static struct	tchars	oldtchars;	/* V7 editing.			*/
static struct	tchars	newtchars = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
static struct	ltchars oldltchars;	/* 4.2 BSD editing.		*/
static struct	ltchars	newltchars = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
static int	modeset = 0;
#ifdef TIOCGWINSZ
static struct winsize win;
#endif

/*
 * Flags telling what parts of the BSD terminal driver structures have been
 * set by ce and need to be reset before exit.  The current set of modes
 * that has been changed is in MODESET.
 */
#define M_SGSET		0x1
#define M_TCSET		0x2
#define M_LTCSET	0x4
#define sgset	(modeset & M_SGSET)
#define tcset	(modeset & M_TCSET)
#define ltcset	(modeset & M_LTCSET)

/*
 * Resize the screen on receipt of a SIGWINCH, assuming that the screen has
 * actually changed its size.  (Window systems have been known to send 
 * spurious resize notices).
 */
#ifdef SIGWINCH
winresize()
{
	refresh(TRUE,1);
} 
#endif

/*
 * This function gets called once, to set up the terminal channel.  
 * Use 8-bit RAW mode, and set all special characters to 0xFF.
 */
ttraw()
{
        extern short ospeed;

	if (ioctl(0, TIOCGETP, (caddr_t) &oldtty) < 0) {
	        mlwrite("ttopen can't get sgtty");
		return(FALSE);
	}
	newtty = oldtty;			/* use structure assignment */
	newtty.sg_flags &= ~(ECHO|CRMOD);	/* Kill echo, CR=>NL.	*/
	newtty.sg_flags |= RAW|ANYP;		/* raw mode for 8 bit path.*/
	if (ioctl(0, TIOCSETP, (caddr_t)  &newtty) < 0) {
	        mlwrite("ttopen can't set sgtty");
		return(FALSE);
	}
	modeset |= M_SGSET;
	ospeed = newtty.sg_ospeed;
	if (ioctl(0, TIOCGETC, (caddr_t) &oldtchars) < 0) {
	        mlwrite("ttopen can't get special chars");
		return(FALSE);
	}
	if (ioctl(0, TIOCSETC, (caddr_t) &newtchars) < 0) {	/* initialized above */
	        mlwrite("ttopen can't set special chars");
		return(FALSE);
	}
	modeset |= M_TCSET;
	if (ioctl(0, TIOCGLTC, (caddr_t) &oldltchars) < 0) {
	        mlwrite("ttopen can't get local chars");
		return(FALSE);
	}
	if (ioctl(0, TIOCSLTC, (caddr_t)  &newltchars) < 0) {	/* initialized above */
	        mlwrite("ttopen can't set local chars");
		return(FALSE);
	}
	modeset |= M_LTCSET;
	ttsetsize();
#ifdef SIGWINCH
	(void)signal(SIGWINCH, winresize);
#endif
	return(TRUE);
}

ttopen()
{
	if (ttraw() == FALSE) 
		panic("\nterminal initialization failure");
}

/*
 * This function gets called just
 * before we go back home to the shell. Put all of
 * the terminal parameters back.
 */
ttcooked()
{
	ttflush();
	if (sgset && (ioctl(0, TIOCSLTC,(caddr_t) &oldltchars) < 0)) {
		mlwrite("ttcooked can't reset local chars");
		return (FALSE);
	}
	modeset &= ~M_SGSET;
	if (tcset && (ioctl(0, TIOCSETC,(caddr_t) &oldtchars) < 0)) {
		mlwrite("ttcooked can't reset special chars");
		return(FALSE);
	}
	modeset &= ~M_TCSET;
	if (ltcset && (ioctl(0, TIOCSETP,(caddr_t) &oldtty) < 0)) {
		mlwrite("ttcooked can't reset sgtty");
		return(FALSE);
	}
	modeset &= ~M_LTCSET;

	return(TRUE);
}

ttclose()
{
	if (ttcooked() == FALSE) 
		panic("");
}

/*
 * Write character to the display.  Characters are buffered up, to make
 * things a little bit more efficient.
 */
ttputc(c)
{
	if (nobuf >= NOBUF)
		ttflush();
	obuf[nobuf++] = c;
}

/*
 * Flush output.
 */
ttflush()
{
	if (nobuf != 0) {
		(void) write(1, obuf, nobuf);
		nobuf = 0;
	}
}

/*
 * Read character from terminal.  All 8 bits are returned, so that you can
 * use a multi-national terminal.
 */
ttgetc()
{
	char	buf[1];

	while (read(0, &buf[0], 1) != 1)
		;
	return (buf[0] & 0xFF);
}

/*
 * Set the number of rows and columns on the terminal via the ioctl 
 * provided by 4.3 BSD (and adopted by SunOS, Ultrix, etc.).  If that
 * fails, or is not defined, we fall back to the termcap entry.  The
 * number of rows is set to the window size - 1 to leave room for the
 * message line, which has its own set of routines.
 */
ttsetsize()
{
	term.t_nrow = term.t_ncol = 0;

#ifdef TIOCGWINSZ
	if (ioctl(0,TIOCGWINSZ,(caddr_t)&win) == 0) {
		term.t_nrow = win.ws_row;
		term.t_ncol = win.ws_col;
	}
#endif

	if (term.t_nrow <= 0)
	    	term.t_nrow = tgetnum("li");
	if (term.t_ncol <= 0)
		term.t_ncol = tgetnum("co");

	if (term.t_nrow <= 0)
		term.t_nrow = 23;
	else
		term.t_nrow--;
	if (term.t_ncol <= 0)
		term.t_ncol = 80;
}

panic(s)
char *s;
{
	ttcooked();
	xpanic(s);
}

typeahead()
{
	int x;

	return ((ioctl(0, FIONREAD, (caddr_t)&x) < 0) ? 0 : x);
}

/*
 * Return a 1 if we can suspend (that is, if the suspend character is defined
 * when we begin processing).
 */

int
cansusp()
{
	return ((oldltchars.t_suspc & 0377) != 0377);
}
